//============================================================================//
//
//  Software Name : MEGADOCK
//
//  Class Name : FFTProcess
//
//  Contact address : Tokyo Institute of Technology, AKIYAMA Lab.
//
//============================================================================//

#include "fft_process.h"

#include <string.h>//shimoda130716

#define NUM_THREADS 512 //should be power of 2

#include <unistd.h>

#ifdef CUFFT
#include <thrust/device_vector.h>
#include <thrust/device_ptr.h>
#include <thrust/sort.h>
#include "conv_kernel.cu"
#endif


//============================================================================//
void FFTProcess::alloc_array(const int &num_fft)
//============================================================================//
{
    //cout << "FFT::alloc_array |" <<num_fft<< endl; cout.flush();
    _Num_fft = num_fft;

    const int     nf3     = _Num_fft * _Num_fft * _Num_fft;
    const int     num_sort  = _parameter->_Num_sort;
    const int     num_angle  = _parameter->_Num_rot_angles;
    const int     no  = _parameter->_Num_output;
    const int nproc2  = _parallel->nproc2();
    int   num_toprank;

    num_toprank = num_angle * num_sort;
    if( no > num_toprank ) num_toprank = no;


    alloc_fft();

    _Select.resize(nproc2);
    for( int i = 0 ; i < nproc2 ; i++ ) {
        _Select[i].resize(num_sort);
    }

    _Top.resize(num_toprank);

    //---------- memory allocation for _Current_rot_angle_num
    _Current_rot_angle_num = new int[nproc2];

    _cputime->record_malloc( sizeof(float)*nf3*2*(1 + nproc2));


    //---------- memory allocation for _FFT_rec_r
    _FFT_rec_r = new float[nf3];
    if( !_FFT_rec_r ) {
        cerr << "[ERROR] Out of memory. Number of listed receptors = ("
             << nf3 << ") for (_FFT_rec_r) in fft_process.cpp!!\n";
        exit(1);
    }

    //---------- memory allocation for _FFT_rec_i
    _FFT_rec_i = new float[nf3];
    if( !_FFT_rec_i ) {
        cerr << "[ERROR] Out of memory. Number of listed receptors = ("
             << nf3 << ") for (_FFT_rec_i) in fft_process.cpp!!\n";
        exit(1);
    }

    //---------- memory allocation for _FFT_lig_r
    _FFT_lig_r = new float*[nproc2];
    for( int i = 0 ; i < nproc2 ; i++ ) {
        _FFT_lig_r[i] = new float[nf3];
        if( !_FFT_lig_r[i] ) {
            cerr << "[ERROR] Out of memory. Number of FFT N*N*N = ("
                 << nf3*nproc2 << ") for (_FFT_lig_r) in fft_process.cpp!!\n";
            exit(1);
        }
    }

    //---------- memory allocation for _FFT_lig_i
    _FFT_lig_i = new float*[nproc2];
    for( int i = 0 ; i < nproc2 ; i++ ) {
        _FFT_lig_i[i] = new float[nf3];
        if( !_FFT_lig_i[i] ) {
            cerr << "[ERROR] Out of memory. Number of FFT N*N*N = ("
                 << nf3*nproc2 << ") for (_FFT_lig_i) in fft_process.cpp!!\n";
            exit(1);
        }
    }

    return;
}

//============================================================================//
void FFTProcess::alloc_fft()
//============================================================================//
{
    const int nf1 = _Num_fft;
    const int nf3 = _Num_fft * _Num_fft * _Num_fft;
    const int num_sort = _parameter->_Num_sort;
    const int nproc2  = _parallel->nproc2();
    const int num_gpu = _parallel->num_gpu();
    int myid2;

#ifdef CUFFT

    //for ligand voxelization on GPU
    const int ng1 = _Num_fft / 2;
    const int ng3 = ng1 * ng1 * ng1;
    const int na = _ligand->num_atoms();
    const int nag = na * ng1;
    const int nThreads = NUM_THREADS;
    const int nBlocks_nf3 = (nf3 + (nThreads-1)) / nThreads;


    _inBuf  = new cufftComplex[nf3];
    _outBuf = new cufftComplex[nf3];

    _cputime->record_malloc( sizeof(cufftComplex)*nf3*2 ); //_in/outBuf

    //printf(" start: %p\n",&_inBuf[0].x);

    /*
    for( int i = 0 ; i < nf3; i++ ) { // This initialization should be executed only once
        if(i<20)printf(" %p %p\n",&_inBuf[i].x,&_inBuf[i].y);
        if(i>nf3-20)printf(" %p %p\n",&_inBuf[i].x,&_inBuf[i].y);
        _inBuf[i] = make_cuComplex(0.0, 0.0);
        _inBuf[i].x = 0.0;
        _inBuf[i].y = 0.0;
    }
    //*/

    int len_inBuf = (int)(((long int)&_inBuf[nf3-1].x) - ((long int)&_inBuf[0].x) + sizeof(_inBuf[nf3-1]))/sizeof(_inBuf[nf3-1]);
    if(len_inBuf !=nf3) printf("# discontinuous memory allocation occurs\n");

    //printf("   end: %ld\n",(long long int)&_inBuf[nf3-1].y - &_inBuf[0].x);

    cufft_plan = new cufftHandle[num_gpu];
    cufft_result = new cufftResult[num_gpu];

    _CUFFTin_dev = new cufftComplex*[num_gpu];
    _CUFFTout_dev = new cufftComplex*[num_gpu];
    _FFT_rec_r_dev = new float*[num_gpu];
    _FFT_rec_i_dev = new float*[num_gpu];

    grid_r_dev = new float*[num_gpu];
    grid_i_dev = new float*[num_gpu];
    grid_coord_dev = new float*[num_gpu];

    radius_core2_dev = new float*[num_gpu];
    radius_surf2_dev = new float*[num_gpu];
    _Charge_dev = new float*[num_gpu];

    xd_dev = new float*[num_gpu];
    yd_dev = new float*[num_gpu];
    zd_dev = new float*[num_gpu];

    mol_coord_dev = new float*[num_gpu];
    mol_coord_orig_dev = new float*[num_gpu];
    mol_coord_orig_center_dev = new float*[num_gpu];
    ligand_rotation_angle_dev = new float*[num_gpu];

    _score_dev = new float*[num_gpu];
    _pos_dev = new int*[num_gpu];

    _score = new float*[num_gpu];
    _pos = new int*[num_gpu];

    //t>1
    temp_score_dev_vec.resize(num_gpu);
    temp_index_dev_vec.resize(num_gpu);

    temp_score_dev_ptr = new float*[num_gpu];
    temp_index_dev_ptr = new int*[num_gpu];

    temp_score_host = new float*[num_gpu];
    temp_index_host = new int*[num_gpu];

    for(int gpuid = 0; gpuid < num_gpu; gpuid++) {
        cudaSetDevice(gpuid);
        cufft_result[gpuid] = cufftPlan3d(&cufft_plan[gpuid], nf1, nf1, nf1, CUFFT_C2C);

        checkCudaErrors( cudaMalloc((void **)&_CUFFTin_dev[gpuid],  sizeof(cufftComplex)*nf3) );
        checkCudaErrors( cudaMalloc((void **)&_CUFFTout_dev[gpuid], sizeof(cufftComplex)*nf3) );
        checkCudaErrors( cudaMalloc((void **)&_FFT_rec_r_dev[gpuid], sizeof(float)*nf3) );
        checkCudaErrors( cudaMalloc((void **)&_FFT_rec_i_dev[gpuid], sizeof(float)*nf3) );

        checkCudaErrors( cudaMalloc((void **)&grid_r_dev[gpuid],  sizeof(float)*ng3));
        checkCudaErrors( cudaMalloc((void **)&grid_i_dev[gpuid],  sizeof(float)*ng3));
        checkCudaErrors( cudaMalloc((void **)&grid_coord_dev[gpuid],  sizeof(float)*ng1));

        checkCudaErrors( cudaMalloc((void **)&radius_core2_dev[gpuid],  sizeof(float)*na));
        checkCudaErrors( cudaMalloc((void **)&radius_surf2_dev[gpuid],  sizeof(float)*na));
        checkCudaErrors( cudaMalloc((void **)&_Charge_dev[gpuid],  sizeof(float)*na));

        checkCudaErrors( cudaMalloc((void **)&xd_dev[gpuid],  sizeof(float)*nag));
        checkCudaErrors( cudaMalloc((void **)&yd_dev[gpuid],  sizeof(float)*nag));
        checkCudaErrors( cudaMalloc((void **)&zd_dev[gpuid],  sizeof(float)*nag));

        checkCudaErrors( cudaMalloc((void **)&mol_coord_dev[gpuid],  sizeof(float)*na*3));
        checkCudaErrors( cudaMalloc((void **)&mol_coord_orig_dev[gpuid],  sizeof(float)*na*3));
        checkCudaErrors( cudaMalloc((void **)&mol_coord_orig_center_dev[gpuid],  sizeof(float)*3));
        checkCudaErrors( cudaMalloc((void **)&ligand_rotation_angle_dev[gpuid],  sizeof(float)*3));

        checkCudaErrors( cudaMalloc((void **)&_score_dev[gpuid], sizeof(float)*nBlocks_nf3*num_sort) );
        checkCudaErrors( cudaMalloc((void **)&_pos_dev[gpuid], sizeof(int)*nBlocks_nf3*num_sort) );

        _score[gpuid] = new float[nBlocks_nf3];
        _pos[gpuid] = new int[nBlocks_nf3];

        if(num_sort!=1) {
            temp_score_dev_vec[gpuid].resize(nf3);
            temp_index_dev_vec[gpuid].resize(nf3);

            temp_score_host[gpuid] = new float[num_sort];
            temp_index_host[gpuid] = new int[num_sort];
        }

    }

    _cputime->record_malloc( sizeof(float)*nBlocks_nf3*num_gpu + sizeof(int)*nBlocks_nf3*num_gpu );

    /*
    size_t devmem_use, devmem_free, devmem_total;
    cudaMemGetInfo(&devmem_free, &devmem_total);
    devmem_use = devmem_total - devmem_free;
    printf(" [GPU memory] Use : %10u (%4.1f%%), Free : %10u (%4.1f%%), Total : %10u\n",devmem_use,(float)(100*devmem_use/devmem_total), devmem_free, (float)(100*devmem_free/devmem_total), devmem_total);
    //*/

#endif /* CUFFT */

    _FFTWin  = (fftwf_complex*) fftwf_malloc(sizeof(fftwf_complex)*nf3*nproc2);
    _FFTWout = (fftwf_complex*) fftwf_malloc(sizeof(fftwf_complex)*nf3*nproc2);

    plan_fftw_forward = new fftwf_plan[nproc2];
    plan_fftw_inverse = new fftwf_plan[nproc2];


    #pragma omp parallel private(myid2) num_threads(nproc2) //limit num of procs
    {
        myid2 = omp_get_thread_num();
        #pragma omp for
        for(int id = 0; id < nproc2; id++) {
            #pragma omp critical
            {
                plan_fftw_forward[myid2]=fftwf_plan_dft_3d(nf1,nf1,nf1,&_FFTWin[nf3*myid2],&_FFTWout[nf3*myid2],FFTW_FORWARD,FFTW_ESTIMATE);
                plan_fftw_inverse[myid2]=fftwf_plan_dft_3d(nf1,nf1,nf1,&_FFTWin[nf3*myid2],&_FFTWout[nf3*myid2],FFTW_BACKWARD,FFTW_ESTIMATE);
            }
        }
    }

    _cputime->record_malloc( sizeof(fftwf_complex)*nf3*2*nproc2 );
    return;
}

//============================================================================//
void FFTProcess::receptor_fft(float *grid_r,float *grid_i)
//============================================================================//
{
    const int num_grid= _Num_fft / 2;
    const int nf3 = _Num_fft * _Num_fft * _Num_fft;
    const int ndata   = ( _Num_fft - num_grid ) / 2;
    const float   theta   = -2.0 * PI / _Num_fft;

    const int num_gpu = _parallel->num_gpu();
    struct timeval et1, et2;
    int myid2;

#ifdef CUFFT
    if(num_gpu > 0) {
        for( int i = 0 ; i < nf3 ; i++ ) {
            _inBuf[i] = make_cuComplex(0.0, 0.0);
        }

        for( int i = 0, m = 0 ; i < num_grid ; i++ ) {
            const int ic = _Num_fft*_Num_fft*(i+ndata);

            for( int j = 0 ; j < num_grid ; j++ ) {
                const int jc = ic + _Num_fft*(j+ndata);

                for( int k = 0 ; k < num_grid ; k++ ) {
                    _inBuf[jc+k+ndata] = make_cuComplex(grid_r[m  ], grid_i[m]);
                    m++;
                }
            }
        }

        cudaSetDevice(0); //_CUFFTin_dev[0] : [0] means 0th GPU

        gettimeofday(&et1,NULL);
        checkCudaErrors( cudaMemcpy(_CUFFTin_dev[0], _inBuf, sizeof(cufftComplex)*nf3, cudaMemcpyHostToDevice) );
        gettimeofday(&et2,NULL);
        _cputime->t4_data_transfer_rec += (et2.tv_sec-et1.tv_sec + (float)((et2.tv_usec-et1.tv_usec)*1e-6));

        fft3d(theta,0); // [0] means performed on 0th GPU

        gettimeofday(&et1,NULL);
        checkCudaErrors( cudaMemcpy(_outBuf,_CUFFTout_dev[0],sizeof(cufftComplex)*nf3,cudaMemcpyDeviceToHost) );
        gettimeofday(&et2,NULL);
        _cputime->t4_data_transfer_rec += (et2.tv_sec-et1.tv_sec + (float)((et2.tv_usec-et1.tv_usec)*1e-6));

        for( int i = 0 ; i < nf3 ; i++ ) {
            _FFT_rec_r[i] = cuCrealf(_outBuf[i]);
            _FFT_rec_i[i] = cuCimagf(_outBuf[i]);
        }

        gettimeofday(&et1,NULL);

        #pragma omp parallel private(myid2) num_threads(num_gpu)
        {
            myid2 = omp_get_thread_num();
            #pragma omp for
            for(int gpuid = 0; gpuid < num_gpu; gpuid++) {
                cudaSetDevice(myid2);
                checkCudaErrors( cudaMemcpy(_FFT_rec_r_dev[myid2], _FFT_rec_r, sizeof(float)*nf3, cudaMemcpyHostToDevice) );
                checkCudaErrors( cudaMemcpy(_FFT_rec_i_dev[myid2], _FFT_rec_i, sizeof(float)*nf3, cudaMemcpyHostToDevice) );
            }
        }

        gettimeofday(&et2,NULL);
        _cputime->t4_data_transfer_rec += (et2.tv_sec-et1.tv_sec + (float)((et2.tv_usec-et1.tv_usec)*1e-6));

    } else {
        for( int i = 0 ; i < nf3 ; i++ ) {
            _FFTWin[i][0] = 0.0;
            _FFTWin[i][1] = 0.0;
        }

        for( int i = 0, m = 0 ; i < num_grid ; i++ ) {
            const int ic = _Num_fft*_Num_fft*(i+ndata);

            for( int j = 0 ; j < num_grid ; j++ ) {
                const int jc = ic + _Num_fft*(j+ndata);

                for( int k = 0 ; k < num_grid ; k++ ) {
                    _FFTWin[jc+k+ndata][0] = grid_r[m  ];
                    _FFTWin[jc+k+ndata][1] = grid_i[m++];
                }
            }
        }

        fft3d(theta,0);

        for( int i = 0 ; i < nf3 ; i++ ) {
            _FFT_rec_r[i] = _FFTWout[i][0];
            _FFT_rec_i[i] = _FFTWout[i][1];
        }
    }

#else
    for( int i = 0 ; i < nf3 ; i++ ) {
        _FFTWin[i][0] = 0.0;
        _FFTWin[i][1] = 0.0;
    }

    for( int i = 0, m = 0 ; i < num_grid ; i++ ) {
        const int ic = _Num_fft*_Num_fft*(i+ndata);

        for( int j = 0 ; j < num_grid ; j++ ) {
            const int jc = ic + _Num_fft*(j+ndata);

            for( int k = 0 ; k < num_grid ; k++ ) {
                _FFTWin[jc+k+ndata][0] = grid_r[m  ];
                _FFTWin[jc+k+ndata][1] = grid_i[m++];
            }
        }
    }

    fft3d(theta,0);

    for( int i = 0 ; i < nf3 ; i++ ) {
        _FFT_rec_r[i] = _FFTWout[i][0];
        _FFT_rec_i[i] = _FFTWout[i][1];
    }
#endif

    return;
}

//============================================================================//
void FFTProcess::ligand_fft(float *grid_r,float *grid_i,int myid2)
//============================================================================//
{
    const int     num_grid = _Num_fft / 2;
    //const int nf1 = _Num_fft;
    //const int nf2 = _Num_fft * _Num_fft;
    const int nf3 = _Num_fft * _Num_fft * _Num_fft;
    const int ndata   = ( _Num_fft - num_grid ) / 2;
    const float   theta   = -2.0 * PI / _Num_fft;
    struct timeval et1, et2;

    gettimeofday(&et1,NULL);
    /*
    for( int i = 0 ; i < nf3 ; i++) {
        //if(i<20)printf("%p %p\n",&_FFTWin[nf3*myid2+i][0],&_FFTWin[nf3*myid2+i][1]);
        //if(i>nf3-20)printf("%p %p\n",&_FFTWin[nf3*myid2+i][0],&_FFTWin[nf3*myid2+i][1]);
        _FFTWin[nf3*myid2+i][0] = 0.0;
        _FFTWin[nf3*myid2+i][1] = 0.0;
    }
    //*/



    gettimeofday(&et2,NULL);
    //printf(" [1-1] %f\n",(et2.tv_sec-et1.tv_sec + (float)((et2.tv_usec-et1.tv_usec)*1e-6)));
    gettimeofday(&et1,NULL);

    memset(_FFTWin[nf3*myid2], 0.0, sizeof(_FFTWin[0])*nf3);


    gettimeofday(&et2,NULL);
    //printf(" [1-2] %f\n",(et2.tv_sec-et1.tv_sec + (float)((et2.tv_usec-et1.tv_usec)*1e-6)));
    gettimeofday(&et1,NULL);

    //for( int a = 0 ; a < 1000 ; a++ ) {

    for( int i = 0, m = 0 ; i < num_grid ; i++ ) {
        const int ic = _Num_fft*_Num_fft*(i+ndata);

        for( int j = 0 ; j < num_grid ; j++ ) {
            int jc = ic + _Num_fft*(j+ndata);

            for( int k = 0 ; k < num_grid ; k++ ) {
                _FFTWin[nf3*myid2+jc+k+ndata][0] = grid_r[m  ];
                _FFTWin[nf3*myid2+jc+k+ndata][1] = grid_i[m++];
            }
        }
    }
    //}

    gettimeofday(&et2,NULL);
    //printf(" [2-1] %f\n",(et2.tv_sec-et1.tv_sec + (float)((et2.tv_usec-et1.tv_usec)*1e-6)));
    gettimeofday(&et1,NULL);
    /*
    const int nloop=1000;

    for( int a = 0 ; a < nloop ; a++ ) {

        for( int i = 0, m = 0 ; i < num_grid ; i++ ) {
            const int ic = nf2*(i+ndata);

            for( int j = 0 ; j < num_grid ; j++ ) {
                int jc = ic + _Num_fft*(j+ndata);

                for( int k = 0, myijk=nf3*myid2+jc+ndata ; k < num_grid ; k++,myijk++ ) {
                    _FFTWin[myijk][0] = grid_r[m  ];
                    _FFTWin[myijk][1] = grid_i[m++];
                }
            }
        }
    }
    //*/

    gettimeofday(&et2,NULL);
    //printf(" [2-2] %f\n",(et2.tv_sec-et1.tv_sec + (float)((et2.tv_usec-et1.tv_usec)*1e-6)));
    gettimeofday(&et1,NULL);

    fft3d(theta,myid2);

    gettimeofday(&et2,NULL);
    //printf(" [3-1] %f\n",(et2.tv_sec-et1.tv_sec + (float)((et2.tv_usec-et1.tv_usec)*1e-6)));
    gettimeofday(&et1,NULL);


    for( int i = 0 ; i < nf3 ; i++ ) {
        _FFT_lig_r[myid2][i] = _FFTWout[nf3*myid2+i][0];
        _FFT_lig_i[myid2][i] = _FFTWout[nf3*myid2+i][1];
    }

    gettimeofday(&et2,NULL);
    //printf(" [4-1] %f\n",(et2.tv_sec-et1.tv_sec + (float)((et2.tv_usec-et1.tv_usec)*1e-6)));
    gettimeofday(&et1,NULL);


    return;
}

//============================================================================//
void FFTProcess::fft_score(int myid2)
//============================================================================//
{
    //const int nf1     = _Num_fft;
    //const int nf2     = nf1*nf1;
    const int nf3     = _Num_fft * _Num_fft * _Num_fft;
    const float   theta   = 2.0 * PI / _Num_fft;
    struct timeval et1, et2;

    if(myid2==0) gettimeofday(&et1,NULL);
    //*
    for( int i = 0 ; i < nf3 ; i++ ) {
        _FFTWin[nf3*myid2+i][0] = _FFT_rec_r[i]*_FFT_lig_r[myid2][i]
                                  + _FFT_rec_i[i]*_FFT_lig_i[myid2][i];
        _FFTWin[nf3*myid2+i][1] = _FFT_rec_r[i]*_FFT_lig_i[myid2][i]
                                  - _FFT_rec_i[i]*_FFT_lig_r[myid2][i];
    }
    //*/

    if(myid2==0) {
        gettimeofday(&et2,NULL);
        _cputime->t3_3_fftprocess_modulation += (et2.tv_sec-et1.tv_sec + (float)((et2.tv_usec-et1.tv_usec)*1e-6));
    }

    if(myid2==0) gettimeofday(&et1,NULL);
    fft3d(theta,myid2);
    if(myid2==0) {
        gettimeofday(&et2,NULL);
        _cputime->t3_4_fftprocess_fft_inverse += (et2.tv_sec-et1.tv_sec + (float)((et2.tv_usec-et1.tv_usec)*1e-6));
    }

    return;
}

//============================================================================//
void FFTProcess::fft3d(const float &theta,int myid2)
//============================================================================//
{
    const int nf1   = _Num_fft;
    //const int nf3 = nf1*nf1*nf1;
    const int nproc2  = _parallel->nproc2();
    const int num_gpu = _parallel->num_gpu();

#ifdef CUFFT

    if(myid2 < num_gpu) {
        cufftHandle plan;
        cufftResult res;

        res = cufftPlan3d(&plan, nf1, nf1, nf1, CUFFT_C2C);
        if(!res == CUFFT_SUCCESS) {
            cout << "!fail to plan 3d FFT (DFT, Lig):" << res << endl;
            exit(-1);
        }

        if( theta < 0.0 ) { //shimdoa130512
            res = cufftExecC2C(plan, &_CUFFTin_dev[myid2][0], &_CUFFTout_dev[myid2][0], CUFFT_FORWARD);
        } else {
            res = cufftExecC2C(plan, &_CUFFTin_dev[myid2][0], &_CUFFTout_dev[myid2][0], CUFFT_INVERSE);
        }

        if(!res == CUFFT_SUCCESS) {
            cout << "!fail to exec 3d FFT(in fft3d()):" << res << endl;
            exit(-1);
        }

        res =  cufftDestroy(plan);

    } else {
        if( theta < 0.0 ) {
            fftwf_execute(plan_fftw_forward[myid2]);
        } else {
            fftwf_execute(plan_fftw_inverse[myid2]);
        }
    }

#else /* FFTW */
    if( theta < 0.0 ) {
        fftwf_execute(plan_fftw_forward[myid2]);
    } else {
        fftwf_execute(plan_fftw_inverse[myid2]);
    }
#endif

    return;
}

//============================================================================//
void FFTProcess::score_sort(int myid2)
//=====================================fft=======================================//
{
    const int num_sort  = _parameter->_Num_sort;
    const int ng1 = _Num_fft / 2;
    const int nf1 = _Num_fft;
    const int nf2 = _Num_fft * _Num_fft;
    const int nf3 = _Num_fft * _Num_fft * _Num_fft;

    for( int i = 0 ; i < num_sort ; i++ ) {
        _Select[myid2][i].score = -99999.0;
    }

    /*
    for( int i = 0 ; i < 3 ; i++ ) {
        for( int j = 0 ; j < 2 ; j++ ) {
            printf(" rec grid_non_zero_region[%d][%d]=%d\n",i,j,_receptor->grid_non_zero_region[i][j]);
        }
    }
    for( int i = 0 ; i < 3 ; i++ ) {
        for( int j = 0 ; j < 2 ; j++ ) {
            printf(" rec grid_zero_space_width[%d][%d]=%d\n",i,j,_receptor->grid_zero_space_width[i][j]);
        }
    }
    for( int i = 0 ; i < 3 ; i++ ) {
        for( int j = 0 ; j < 2 ; j++ ) {
            printf(" lig grid_non_zero_region[%d][%d]=%d\n",i,j,_ligand->grid_non_zero_region[i][j]);
        }
    }
    for( int i = 0 ; i < 3 ; i++ ) {
        for( int j = 0 ; j < 2 ; j++ ) {
            printf(" lig grid_zero_space_width[%d][%d]=%d\n",i,j,_ligand->grid_zero_space_width[i][j]);
        }
    }
    //*/

    int fft_zero_region[3][2]= {0};
    int max_search_scope[3][2]= {0};
    for( int i = 0 ; i < 3 ; i++ ) {
        fft_zero_region[i][0]= ng1 - (_receptor->grid_zero_space_width[i][0] + _ligand->grid_zero_space_width[i][1]);
        fft_zero_region[i][1]= ng1 + (_receptor->grid_zero_space_width[i][1] + _ligand->grid_zero_space_width[i][0]) + 1;
        //printf(" fft_zero_region[%d]=%d %d\n",i,fft_zero_region[i][0],fft_zero_region[i][1]);
    }

    /*
    for( int ijk = 0; ijk < nf3 ; ijk++ ) {
        int i = ijk / nf2;
        int j = (ijk / _Num_fft) % _Num_fft;
        int k = ijk % _Num_fft;
        if(
            !(
                (i<fft_zero_region[0][0] ||
                 fft_zero_region[0][1]<i ) &&
                (j<fft_zero_region[1][0] ||
                 fft_zero_region[1][1]<j ) &&
                (k<fft_zero_region[2][0] ||
                 fft_zero_region[2][1]<k )
            ) &&
            (_FFTWout[nf3*myid2+ijk][0] / nf3 > 1.0)
        ) {
            printf(" %3d*%3d*%3d : %f\n",i,j,k,_FFTWout[nf3*myid2+ijk][0] / nf3);
        }
    }
    //*/

    struct timeval et1, et2;

    gettimeofday(&et1,NULL);

    for( int ii = 0 ; ii < 1 ; ii++ ) {
        if(_parameter->tem_flag3 == 1) {
            if(num_sort!=1) {
                const int mynf3= nf3*myid2;
                for( int block = 0 ; block < 8 ; block++) {
                    max_search_scope[0][0]=(block%2==0)?0:fft_zero_region[0][1];
                    max_search_scope[0][1]=(block%2==0)?fft_zero_region[0][0]:nf1;
                    max_search_scope[1][0]=(block/2%2==0)?0:fft_zero_region[1][1];
                    max_search_scope[1][1]=(block/2%2==0)?fft_zero_region[1][0]:nf1;
                    max_search_scope[2][0]=(block/4==0)?0:fft_zero_region[2][1];
                    max_search_scope[2][1]=(block/4==0)?fft_zero_region[2][0]:nf1;
                    //printf(" (%3d-%3d)*(%3d-%3d)*(%3d-%3d)\n",max_search_scope[0][0],max_search_scope[0][1],max_search_scope[1][0],max_search_scope[1][1],max_search_scope[2][0],max_search_scope[2][1]);
                    for( int i = max_search_scope[0][0], ic=max_search_scope[0][0]*nf2 ; i < max_search_scope[0][1] ; i++,ic+=nf2 ) {
                        for( int j = max_search_scope[1][0], jc=ic+max_search_scope[1][0]*nf1 ; j < max_search_scope[1][1] ; j++,jc+=nf1 ) {
                            for( int k = max_search_scope[2][0], ijk=jc+max_search_scope[2][0], myijk=mynf3+jc+max_search_scope[2][0] ; k < max_search_scope[2][1] ; k++,ijk++,myijk++ ) {
                                const float raw = _FFTWout[myijk][0] / nf3;
                                if( raw < _Select[myid2][num_sort-1].score) continue;
                                for( int l = 0 ; l < num_sort ; l++ ) {
                                    if( raw > _Select[myid2][l].score ) {
                                        for( int m = num_sort-1 ; m > l ; m-- ) {
                                            _Select[myid2][m] = _Select[myid2][m-1];
                                        }
                                        _Select[myid2][l].score    = raw;
                                        _Select[myid2][l].index[1] = ijk / nf2;
                                        _Select[myid2][l].index[2] = (ijk / _Num_fft) % _Num_fft;
                                        _Select[myid2][l].index[3] = ijk % _Num_fft;
                                        break;
                                    }
                                }
                            }
                        }
                    }
                }
            } else { // num_sort = 1, select only 1 score per 1 ligand angle
                float temp_top_score = 0.0;
                int temp_top_pos = 0;
                const int mynf3= nf3*myid2;
                for( int block = 0 ; block < 8 ; block++) {
                    max_search_scope[0][0]=(block%2==0)?0:fft_zero_region[0][1];
                    max_search_scope[0][1]=(block%2==0)?fft_zero_region[0][0]:nf1;
                    max_search_scope[1][0]=(block/2%2==0)?0:fft_zero_region[1][1];
                    max_search_scope[1][1]=(block/2%2==0)?fft_zero_region[1][0]:nf1;
                    max_search_scope[2][0]=(block/4==0)?0:fft_zero_region[2][1];
                    max_search_scope[2][1]=(block/4==0)?fft_zero_region[2][0]:nf1;
                    //printf(" (%d-%d)*(%d-%d)*(%d-%d)\n",max_search_scope[0][0],max_search_scope[0][1],max_search_scope[1][0],max_search_scope[1][1],max_search_scope[2][0],max_search_scope[2][1]);
                    for( int i = max_search_scope[0][0], ic=max_search_scope[0][0]*nf2 ; i < max_search_scope[0][1] ; i++,ic+=nf2 ) {
                        for( int j = max_search_scope[1][0], jc=ic+max_search_scope[1][0]*nf1 ; j < max_search_scope[1][1] ; j++,jc+=nf1 ) {
                            for( int k = max_search_scope[2][0], ijk=jc, myijk=mynf3+jc ; k < max_search_scope[2][1] ; k++,ijk++,myijk++ ) {
                                const float raw = _FFTWout[myijk][0];
                                if (temp_top_score < raw) {
                                    temp_top_score = raw;
                                    temp_top_pos = ijk;
                                }
                            }
                        }
                    }
                }
                _Select[myid2][0].score    = temp_top_score / nf3;
                _Select[myid2][0].index[1] = temp_top_pos / nf2;
                _Select[myid2][0].index[2] = (temp_top_pos / _Num_fft) % _Num_fft;
                _Select[myid2][0].index[3] = temp_top_pos % _Num_fft;
            }
        } else {
            if(num_sort!=1) {
                for( int i = 0,myi= nf3*myid2 ; i < nf3 ; i++,myi++ ) {
                    const float raw = _FFTWout[myi][0] / nf3;
                    if( raw < _Select[myid2][num_sort-1].score) continue;
                    for( int j = 0 ; j < num_sort ; j++ ) {
                        if( raw > _Select[myid2][j].score ) {
                            for( int k = num_sort-1 ; k > j ; k-- ) {
                                _Select[myid2][k] = _Select[myid2][k-1];
                            }
                            _Select[myid2][j].score    = raw;
                            _Select[myid2][j].index[1] = i / nf2;
                            _Select[myid2][j].index[2] = (i / _Num_fft) % _Num_fft;
                            _Select[myid2][j].index[3] = i % _Num_fft;
                            break;
                        }
                    }
                }
            } else { // num_sort = 1, take only 1 score per angle
                float temp_top_score = 0.0;
                int temp_top_pos = 0;
                for( int i = 0, myi=nf3*myid2 ; i < nf3 ; i++,myi++ ) {
                    const float raw = _FFTWout[myi][0];
                    if (temp_top_score < raw) {
                        temp_top_score = raw;
                        temp_top_pos = i;
                    }
                }
                _Select[myid2][0].score    = temp_top_score / nf3;
                _Select[myid2][0].index[1] = temp_top_pos / nf2;
                _Select[myid2][0].index[2] = (temp_top_pos / _Num_fft) % _Num_fft;
                _Select[myid2][0].index[3] = temp_top_pos % _Num_fft;
            }
        }
    }

    gettimeofday(&et2,NULL);
    /*
    int nf3rev = (nf1 - (fft_zero_region[0][1] - fft_zero_region[0][0])) *
                 (nf1 - (fft_zero_region[1][1] - fft_zero_region[1][0])) *
                 (nf1 - (fft_zero_region[2][1] - fft_zero_region[2][0]));


    if (_parameter->tem_flag3 != 1) nf3rev = nf3;
    //*/
    //printf(" %8d/%8d [grid], %4.1f[%%], %2d [t], %10.6f [sec]\n",nf3rev,nf3,(float)(nf3rev*100.0/nf3),num_sort,1.0*(et2.tv_sec-et1.tv_sec + (float)((et2.tv_usec-et1.tv_usec)*1e-6)));

    for( int i = 0 ; i < num_sort ; i++ ) {
        //printf(" top %d %f\n",i,_Select[myid2][i].score);
        _Select[myid2][i].index[0] = _Current_rot_angle_num[myid2];
    }

    for( int i = 0 ; i < num_sort ; i++ ) {
        _Top[_Current_rot_angle_num[myid2]*num_sort+i] = _Select[myid2][i];
    }

    return;
}


//============================================================================//
bool ascending_struct(const SortScore &s1,const SortScore &s2)  // for sort
//============================================================================//
{
    return s1.score > s2.score;
}

//============================================================================//
void FFTProcess::sort_index(float *fwork,int *iwork)
//============================================================================//
{
    const int     no  = _parameter->_Num_output;
    const int     nt  = (_Top.size() < no) ? _Top.size() : no;

    partial_sort(_Top.begin(),_Top.begin()+nt,_Top.end(),
                 ascending_struct);

    return;
}

//============================================================================//
void FFTProcess::top_score_clean()
//============================================================================//
{
    const int     num_sort  = _parameter->_Num_sort;
    const int     num_angle  = _parameter->_Num_rot_angles;
    const int     no  = _parameter->_Num_output;
    int   num_toprank;

    num_toprank = num_angle * num_sort;
    if( no > num_toprank ) num_toprank = no;

    for( int j = 0 ; j < num_toprank ; j++ ) {
        _Top[j].score = 0.0;
    }

    return;
}

#ifdef CUFFT
//============================================================================//
void FFTProcess::cuda_fft(float *grid_r,float *grid_i,float *grid_coord,float *mol_coord,float *theta,int myid2)
//============================================================================//
{
    const int nf1 = _Num_fft;
    const int nf2 = nf1 * nf1;
    const int nf3 = nf2 * nf1;

    const int num_sort = _parameter->_Num_sort;
    const int na = _ligand->num_atoms();

    struct timeval et1, et2;
    struct timeval et3, et4;
    if(myid2==0) gettimeofday(&et1,NULL);

    float temp_top_score = -999999.0;
    int temp_top_pos = -999999;

    const int nThreads = NUM_THREADS;
    const int nBlocks_nf3 = (nf3 + (nThreads-1)) / nThreads;
    if(nBlocks_nf3 * nThreads < nf3) {
        printf(" nf3:%d, nBlocks_nf3:%d, nThreads:%d , nf3=nBlocks_nf3*nThreads\n",nf3,nBlocks_nf3,nThreads);
        fprintf(stderr, " [ERROR] too large FFT size. nf3:%d, nBlocks_nf3:%d\n", nf3, nBlocks_nf3);
        exit(1);
    }

    cudaSetDevice(myid2);
    //printf(" #p10 [myid=%d]\n",myid2);

    ligand_voxelization_on_dev(theta,myid2);
    checkCudaErrors( cudaDeviceSynchronize() );

    if(myid2==0) gettimeofday(&et2,NULL);
    if(myid2==0) _cputime->t3_1_ligand_voxelization += (et2.tv_sec-et1.tv_sec + (float)((et2.tv_usec-et1.tv_usec)*1e-6));
    if(myid2==0) gettimeofday(&et1,NULL);

    cufft_result[myid2] = cufftExecC2C(cufft_plan[myid2], &_CUFFTin_dev[myid2][0], &_CUFFTout_dev[myid2][0], CUFFT_FORWARD);
    if(!cufft_result[myid2] == CUFFT_SUCCESS) {
        cout << "!fail to exec 3d FFT (DFT, Lig):" << cufft_result[myid2] << endl;
        exit(-1);
    }
    //*/
    checkCudaErrors( cudaDeviceSynchronize() );

    if(myid2==0) gettimeofday(&et2,NULL);
    if(myid2==0) _cputime->t3_2_fftprocess_ligand_fft += (et2.tv_sec-et1.tv_sec + (float)((et2.tv_usec-et1.tv_usec)*1e-6));

    if(myid2==0) gettimeofday(&et1,NULL);
    modulation<<<nBlocks_nf3, nThreads>>>(nf3, _FFT_rec_r_dev[myid2], _FFT_rec_i_dev[myid2], _CUFFTout_dev[myid2], _CUFFTin_dev[myid2]);

    checkCudaErrors( cudaDeviceSynchronize() );

    if(myid2==0) gettimeofday(&et2,NULL);
    if(myid2==0) _cputime->t3_3_fftprocess_modulation += (et2.tv_sec-et1.tv_sec + (float)((et2.tv_usec-et1.tv_usec)*1e-6));
    if(myid2==0) gettimeofday(&et1,NULL);

    cufft_result[myid2] = cufftExecC2C(cufft_plan[myid2], &_CUFFTin_dev[myid2][0], &_CUFFTout_dev[myid2][0], CUFFT_INVERSE);
    if(!(cufft_result[myid2] == CUFFT_SUCCESS)) {
        cout << "!fail to plan 3d FFT (IDFT):" << cufft_result[myid2] << endl;
        exit(-1);
    }
    //*
    checkCudaErrors( cudaDeviceSynchronize() );
    if(myid2==0) gettimeofday(&et2,NULL);
    if(myid2==0) _cputime->t3_4_fftprocess_fft_inverse += (et2.tv_sec-et1.tv_sec + (float)((et2.tv_usec-et1.tv_usec)*1e-6));
    if(myid2==0) gettimeofday(&et1,NULL);

    // Search max score translation position from _CUFFTout_dev[nf3]

    //printf(" t=%d per angle\n",num_sort);

    for( int i = 0 ; i < num_sort ; i++ ) {
        _Select[myid2][i].score = -99999.0;
    }

    if(num_sort!=1) {
        temp_score_dev_ptr[myid2] = thrust::raw_pointer_cast(&temp_score_dev_vec[myid2][0]);
        temp_index_dev_ptr[myid2] = thrust::raw_pointer_cast(&temp_index_dev_vec[myid2][0]);

        max_pos_multi_set<<<nBlocks_nf3, nThreads>>>(nf3, _CUFFTout_dev[myid2],temp_score_dev_ptr[myid2],temp_index_dev_ptr[myid2]);

        thrust::sort_by_key(temp_score_dev_vec[myid2].begin(), temp_score_dev_vec[myid2].end(), temp_index_dev_vec[myid2].begin(), thrust::greater<float>());

        checkCudaErrors( cudaMemcpy(temp_score_host[myid2],temp_score_dev_ptr[myid2],sizeof(float)*num_sort,cudaMemcpyDeviceToHost) );
        checkCudaErrors( cudaMemcpy(temp_index_host[myid2],temp_index_dev_ptr[myid2],sizeof(int)*num_sort,cudaMemcpyDeviceToHost) );

        for( int i = 0 ; i < num_sort ; i++ ) {
            //printf(" RESULT[%d]: %d, %f\n",i,temp_index_host[myid2][i],temp_score_host[myid2][i]/nf3);
            _Select[myid2][i].score    = temp_score_host[myid2][i] / nf3;
            _Select[myid2][i].index[1] = temp_index_host[myid2][i] / nf2;
            _Select[myid2][i].index[2] = (temp_index_host[myid2][i] / nf1) % nf1;
            _Select[myid2][i].index[3] = temp_index_host[myid2][i] % nf1;
        }

    } else { // num_sort = 1, select only 1 score per 1 ligand angle
        max_pos_single<<<nBlocks_nf3, nThreads, sizeof(float)*nThreads>>>(nf3, _CUFFTout_dev[myid2],  _score_dev[myid2], _pos_dev[myid2]);
        checkCudaErrors( cudaDeviceSynchronize() );

        if(myid2==0) gettimeofday(&et3,NULL);
        checkCudaErrors( cudaMemcpy(_score[myid2],_score_dev[myid2],sizeof(float)*nBlocks_nf3,cudaMemcpyDeviceToHost) );
        checkCudaErrors( cudaMemcpy(_pos[myid2],_pos_dev[myid2],sizeof(int)*nBlocks_nf3,cudaMemcpyDeviceToHost) );
        if(myid2==0) gettimeofday(&et4,NULL);
        if(myid2==0) _cputime->t4_data_transfer_in_loop += (et4.tv_sec-et3.tv_sec + (float)((et4.tv_usec-et3.tv_usec)*1e-6));
        checkCudaErrors( cudaDeviceSynchronize() );

        for(int i=0; i<nBlocks_nf3; i++) {
            if(temp_top_score < _score[myid2][i]) {
                temp_top_score = _score[myid2][i];
                temp_top_pos = _pos[myid2][i];
            }
        }

        //printf("  m:%f\n\n",temp_top_score);
        //printf("%g (%d) [%d %d %d]\n", temp_top_score, _p, temp_top_pos/(n*n),(temp_top_pos/n)%n, temp_top_pos%n );
        //printf("<%d> %g (%d/%d) %d\n", nBlocks,temp_top_score, temp_top_pos, nf3, temp_top_pos/nf2);

        _Select[myid2][0].score    = temp_top_score;
        _Select[myid2][0].index[1] = temp_top_pos / nf2;
        _Select[myid2][0].index[2] = (temp_top_pos / nf1) % nf1;
        _Select[myid2][0].index[3] = temp_top_pos % nf1;
    }

    //*** score_sort ***********************************************************

    for( int i = 0 ; i < num_sort ; i++ ) {
        _Select[myid2][i].index[0] = _Current_rot_angle_num[myid2];
        _Top[_Current_rot_angle_num[myid2]*num_sort+i] = _Select[myid2][i];
    }

    checkCudaErrors( cudaDeviceSynchronize() );
    if(myid2==0) gettimeofday(&et2,NULL);
    if(myid2==0) _cputime->t3_5_fftprocess_score_sort += (et2.tv_sec-et1.tv_sec + (float)((et2.tv_usec-et1.tv_usec)*1e-6));

    return;
}


//============================================================================//
void FFTProcess::ligand_voxelization_on_dev(float *theta,int myid2)
//============================================================================//
{
    const int ng1 = _Num_fft / 2;
    const int ng3 = ng1 * ng1 * ng1;
    const int nf1 = _Num_fft;
    const int nf2 = nf1 * nf1;
    const int nf3 = nf2 * nf1;

    const float delta = 1.0;
    const float surface = 1.0;
    const float grid_width = 1.2;

    const int na = _ligand->num_atoms();
    const int nag = na * ng1;

    struct timeval et1, et2;
    struct timeval et3, et4;

    const int nThreads = NUM_THREADS;
    const int nBlocks_na = (na + (nThreads-1)) / nThreads;
    const int nBlocks_nag = (nag + (nThreads-1)) / nThreads;
    const int nBlocks_ng3 = (ng3 + (nThreads-1)) / nThreads;
    const int nBlocks_nf3 = (nf3 + (nThreads-1)) / nThreads;
    if(nBlocks_nf3 * nThreads < nf3) {
        printf(" nf3:%d, nBlocks_nf3:%d, nThreads:%d , nf3=nBlocks_nf3*nThreads\n",nf3,nBlocks_nf3,nThreads);
        fprintf(stderr, " [ERROR] too large FFT size. nf3:%d, nBlocks_nf3:%d\n", nf3, nBlocks_nf3);
        exit(1);
    }

    //*
    //transfer ligand angle & calc xd,yd,zd,mol_coord
    if(myid2==0) gettimeofday(&et3,NULL);

    if(myid2==0) gettimeofday(&et1,NULL);
    checkCudaErrors( cudaMemcpy(ligand_rotation_angle_dev[myid2], theta, sizeof(float)*3, cudaMemcpyHostToDevice) );
    if(myid2==0) gettimeofday(&et2,NULL);
    if(myid2==0) _cputime->t3_1_ligand_voxelization += (et2.tv_sec-et1.tv_sec + (float)((et2.tv_usec-et1.tv_usec)*1e-6));

    lig_rotation<<<nBlocks_na, nThreads>>>(na, ligand_rotation_angle_dev[myid2],mol_coord_orig_dev[myid2], mol_coord_orig_center_dev[myid2], mol_coord_dev[myid2]);
    lig_calc_dis_atomgrid<<<nBlocks_nag, nThreads>>>(na, ng1, xd_dev[myid2], yd_dev[myid2], zd_dev[myid2], grid_coord_dev[myid2], mol_coord_dev[myid2]);
    checkCudaErrors( cudaDeviceSynchronize() );
    if(myid2==0) gettimeofday(&et4,NULL);
    if(myid2==0) _cputime->t3_1_1_ligvoxgpu_copy_htod += (et4.tv_sec-et3.tv_sec + (float)((et4.tv_usec-et3.tv_usec)*1e-6));

    //grid[] initialize
    if(myid2==0) gettimeofday(&et3,NULL);
    lig_vox_init_grid<<<nBlocks_ng3, nThreads>>>(ng3,grid_r_dev[myid2],grid_i_dev[myid2]);
    lig_vox_init_fft<<<nBlocks_nf3, nThreads>>>(nf3,_CUFFTin_dev[myid2]);
    checkCudaErrors( cudaDeviceSynchronize() );
    if(myid2==0) gettimeofday(&et4,NULL);
    if(myid2==0) _cputime->t3_1_2_ligvoxgpu_kernel_init += (et4.tv_sec-et3.tv_sec + (float)((et4.tv_usec-et3.tv_usec)*1e-6));

    //atom fill(core)
    if(myid2==0) gettimeofday(&et3,NULL);
    lig_vox_fill<<<nBlocks_ng3, nThreads>>>
    (ng1,na,delta,radius_core2_dev[myid2],xd_dev[myid2],yd_dev[myid2],zd_dev[myid2],grid_coord_dev[myid2],mol_coord_dev[myid2],grid_r_dev[myid2]);
    checkCudaErrors( cudaDeviceSynchronize() );
    if(myid2==0) gettimeofday(&et4,NULL);
    if(myid2==0) _cputime->t3_1_3_ligvoxgpu_kernel_fill_core += (et4.tv_sec-et3.tv_sec + (float)((et4.tv_usec-et3.tv_usec)*1e-6));

    //surface cutting
    if(myid2==0) gettimeofday(&et3,NULL);
    lig_vox_surface_cut_CtoT<<<nBlocks_ng3, nThreads>>>(ng1,delta,grid_r_dev[myid2]);
    checkCudaErrors( cudaDeviceSynchronize() );
    if(myid2==0) gettimeofday(&et4,NULL);
    if(myid2==0) _cputime->t3_1_4_ligvoxgpu_kernel_cut_surf += (et4.tv_sec-et3.tv_sec + (float)((et4.tv_usec-et3.tv_usec)*1e-6));

    //atom fill(surf)
    if(myid2==0) gettimeofday(&et3,NULL);
    lig_vox_fill<<<nBlocks_ng3, nThreads>>>
    (ng1,na,surface,radius_surf2_dev[myid2],xd_dev[myid2],yd_dev[myid2],zd_dev[myid2],grid_coord_dev[myid2],mol_coord_dev[myid2],grid_r_dev[myid2]);
    checkCudaErrors( cudaDeviceSynchronize() );
    if(myid2==0) gettimeofday(&et4,NULL);
    if(myid2==0) _cputime->t3_1_5_ligvoxgpu_kernel_fill_surf += (et4.tv_sec-et3.tv_sec + (float)((et4.tv_usec-et3.tv_usec)*1e-6));

    //electro
    if(myid2==0) gettimeofday(&et3,NULL);

    if(_parameter->lig_elec_serial_flag == 0) {
        lig_vox_elec<<<nBlocks_ng3, nThreads>>>(ng1, na, grid_width, _Charge_dev[myid2], mol_coord_dev[myid2], grid_i_dev[myid2]);
    } else {
        lig_vox_elec_serial<<<nBlocks_ng3, nThreads>>>(ng1, na, grid_width, _Charge_dev[myid2], mol_coord_dev[myid2], grid_i_dev[myid2]);
    }

    /*
    float *tem_grid;
    const int ng2=ng1*ng1;
    tem_grid = new float[ng3];
    checkCudaErrors( cudaMemcpy(tem_grid, grid_i_dev[myid2], sizeof(float)*ng3, cudaMemcpyDeviceToHost) );
    //for(int i=0;i<ng3;i++) if(tem_grid[i]!=0.0) printf(" [%03d,%03d,%03d] :  %6.3f\n",i/ng2,(i/ng1)%ng1,i%ng1,tem_grid[i]);
    //*/

    checkCudaErrors( cudaDeviceSynchronize() );
    if(myid2==0) gettimeofday(&et4,NULL);
    if(myid2==0) _cputime->t3_1_6_ligvoxgpu_kernel_elec += (et4.tv_sec-et3.tv_sec + (float)((et4.tv_usec-et3.tv_usec)*1e-6));

    //set Voxel grid[ng3] into center of FFT grid[nf3]
    if(myid2==0) gettimeofday(&et3,NULL);
    ligand_voxel_set<<<nBlocks_ng3, nThreads>>>(ng1,_CUFFTin_dev[myid2],grid_r_dev[myid2],grid_i_dev[myid2]);
    checkCudaErrors( cudaDeviceSynchronize() );
    if(myid2==0) gettimeofday(&et4,NULL);
    if(myid2==0) _cputime->t3_1_7_ligvoxgpu_kernel_set_array += (et4.tv_sec-et3.tv_sec + (float)((et4.tv_usec-et3.tv_usec)*1e-6));

    //*/
}


//============================================================================//
void FFTProcess::ligand_info_transfer(float *grid_coord)
//============================================================================//
{
    const int ng1 = _Num_fft / 2;
    const int na = _ligand->num_atoms();
    const int num_gpu = _parallel->num_gpu();
    const float   rcore2 = 1.5;           // ZDOCK parameter
    const float   rsurf2 = 1.0;           // ZDOCK parameter
    struct timeval et1, et2;

    float *radius_core2;
    float *radius_surf2;
    radius_core2 = new float[na];
    radius_surf2 = new float[na];

    for(int i=0; i<na; i++) {
        radius_core2[i] = _ligand->_Radius[i] * _ligand->_Radius[i] * rcore2;
        radius_surf2[i] = _ligand->_Radius[i] * _ligand->_Radius[i] * rsurf2;
    }

    gettimeofday(&et1,NULL);
    for(int gpuid = 0; gpuid < num_gpu; gpuid++) {
        cudaSetDevice(gpuid);
        checkCudaErrors( cudaMemcpy(radius_core2_dev[gpuid], radius_core2, sizeof(float)*na, cudaMemcpyHostToDevice) );
        checkCudaErrors( cudaMemcpy(radius_surf2_dev[gpuid], radius_surf2, sizeof(float)*na, cudaMemcpyHostToDevice) );
        checkCudaErrors( cudaMemcpy(_Charge_dev[gpuid], _ligand->_Charge, sizeof(float)*na, cudaMemcpyHostToDevice) );
        checkCudaErrors( cudaMemcpy(grid_coord_dev[gpuid], grid_coord, sizeof(float)*ng1, cudaMemcpyHostToDevice) );
        checkCudaErrors( cudaMemcpy(mol_coord_orig_dev[gpuid], _ligand->_Coordinate, sizeof(float)*na*3, cudaMemcpyHostToDevice) );
        checkCudaErrors( cudaMemcpy(mol_coord_orig_center_dev[gpuid], _ligand->_Center, sizeof(float)*3, cudaMemcpyHostToDevice) );
    }

    gettimeofday(&et2,NULL);
    _cputime->t4_data_transfer_lig += (et2.tv_sec-et1.tv_sec + (float)((et2.tv_usec-et1.tv_usec)*1e-6));

    delete[] radius_core2;
    delete[] radius_surf2;

    return;
}

#endif /* CUFFT */



//============================================================================//
void FFTProcess::fft_memory_free()
//============================================================================//
{
    const int nproc2    = _parallel->nproc2();
    const int num_gpu = _parallel->num_gpu();
    const int nf3 = _Num_fft * _Num_fft * _Num_fft;
    const int num_sort = _parameter->_Num_sort;

    for(int id = 0; id < nproc2; id++) {
        fftwf_destroy_plan(plan_fftw_forward[id]);
        fftwf_destroy_plan(plan_fftw_inverse[id]);
    }

    for( int i = 0 ; i < nproc2 ; i++ ) {
        delete [] _FFT_lig_r[i];
        delete [] _FFT_lig_i[i];
    }
    _cputime->record_free( sizeof(float)*nf3*2*(nproc2));

#ifdef CUFFT

    const int nThreads = NUM_THREADS;
    const int nBlocks_nf3 = (nf3 + (nThreads-1)) / nThreads;

    for(int gpuid = 0; gpuid < num_gpu; gpuid++) {
        cudaSetDevice(gpuid);

        cufftDestroy(cufft_plan[gpuid]);

        checkCudaErrors( cudaFree(_CUFFTin_dev[gpuid]));
        checkCudaErrors( cudaFree(_CUFFTout_dev[gpuid]));
        checkCudaErrors( cudaFree(_FFT_rec_r_dev[gpuid]));
        checkCudaErrors( cudaFree(_FFT_rec_i_dev[gpuid]));

        checkCudaErrors( cudaFree(grid_r_dev[gpuid]));
        checkCudaErrors( cudaFree(grid_i_dev[gpuid]));
        checkCudaErrors( cudaFree(grid_coord_dev[gpuid]));

        checkCudaErrors( cudaFree(radius_core2_dev[gpuid]));
        checkCudaErrors( cudaFree(radius_surf2_dev[gpuid]));
        checkCudaErrors( cudaFree(_Charge_dev[gpuid]));

        checkCudaErrors( cudaFree(xd_dev[gpuid]));
        checkCudaErrors( cudaFree(yd_dev[gpuid]));

        checkCudaErrors( cudaFree(zd_dev[gpuid]));

        checkCudaErrors( cudaFree(mol_coord_dev[gpuid]));
        checkCudaErrors( cudaFree(mol_coord_orig_dev[gpuid]));
        checkCudaErrors( cudaFree(mol_coord_orig_center_dev[gpuid]));
        checkCudaErrors( cudaFree(ligand_rotation_angle_dev[gpuid]));

        checkCudaErrors( cudaFree(_score_dev[gpuid]));
        checkCudaErrors( cudaFree(_pos_dev[gpuid]));

        delete [] _score[gpuid];
        delete [] _pos[gpuid];

        if(num_sort!=1) {
            delete [] temp_score_host[gpuid];
            delete [] temp_index_host[gpuid];
        }
    }

    _cputime->record_free( sizeof(float)*nBlocks_nf3*num_gpu + sizeof(int)*nBlocks_nf3*num_gpu );


#endif

    return;
}
